
package de.grosenick.brickapi;

/**
 * Abstract {@link Command} that implements the {@link #execute(BrickConnector)} method
 * in a way that a "called" flag is set which can be queried using {@link #isInQueue()}
 * afterwards.
 * <P>
 * Derived classes must implement the method {@link #doExecute(BrickConnector)}
 * instead.
 * 
 * @author Holger Grosenick
 */
public abstract class AbstractCommand implements Command
{
    private Status status = Status.IDLE;
    
    /** set to 0 at creation time, incremented with each call to {@link #execute(BrickConnector)} */
    protected int callCount = 0;
    protected boolean debugOn;

    /* (non-Javadoc)
     * @see de.grosenick.brickapi.Command#execute(de.grosenick.brickapi.BrickConnector)
     */
    @Override
    public synchronized final void execute(BrickConnector brickConnector)
        throws Exception
    {
        status = Status.ACTIVE;
        try {
            doExecute(brickConnector);
        }
        finally {
            // be sure to have a clean state in case of exception
            ++callCount;
            status = Status.IDLE;
        }
    }

    /**
     * Similar as {@link Command#execute(BrickConnector)}, but has to be implemented by the
     * derived class.
     * 
     * @param brickConnector the connector to use
     * @throws Exception in case of any error
     */
    protected abstract void doExecute(BrickConnector brickConnector)
        throws Exception;

    /* (non-Javadoc)
     * @see de.grosenick.brickapi.Command#enqueue(java.util.AbstractQueue)
     */
    @Override
    public synchronized final void enqueue(java.util.AbstractQueue<Command> q)
    {
        if (status != Status.IDLE) {
            throw new IllegalStateException("The command object is not idle!");
        }
        status = Status.IN_QUEUE;
        q.add(this);
    }

    /* (non-Javadoc)
     * @see de.grosenick.brickapi.Command#isInQueue()
     */
    @Override
    public final boolean isInQueue() {
        return status == Status.IN_QUEUE;
    }

    /* (non-Javadoc)
     * @see de.grosenick.brickapi.Command#isActive()
     */
    @Override
    public final boolean isActive() {
        return status == Status.ACTIVE;
    }

    /* (non-Javadoc)
     * @see de.grosenick.brickapi.Command#isIdle()
     */
    @Override
    public final boolean isIdle() {
        return status == Status.IDLE;
    }

    /**
     * Always returns null by default.
     * @see de.grosenick.brickapi.Command#getResult()
     */
    @Override
    public Object getResult() {
        return null;
    }
    
    /* (non-Javadoc)
     * @see de.grosenick.brickapi.Command#getStatus()
     */
    @Override
    public Status getStatus() {
        return status;
    }

    /**
     * Returns how often this command was sent to the brick(let).
     * @return an int
     */
    public int getCallCount() {
        return callCount;
    }
    
    /**
     * Sets a debug flag to on or off which might cause some traces to get active.
     * @param d true or false
     */
    public void setDebugOn(boolean d) {
        debugOn = d;
    }
    
    @Override
    public Object clone()
    {
        try {
            AbstractCommand o = (AbstractCommand) super.clone();
            o.status = Status.IDLE;
            return o;
        }
        catch (CloneNotSupportedException e) {
            e.printStackTrace();
            return null;
        }
    }
}
