package de.grosenick.brickapi;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import com.tinkerforge.*;
import com.tinkerforge.IPConnection.EnumerateListener;


/**
 * Main class that is used to connect to bricks and bricklets.
 * It enumerates the devices in the constructor and can directly
 * access some devices if those are unique.
 * 
 * @author Holger Grosenick
 */
public class BrickConnector implements EnumerateListener
{
    // Most commonly used bricks
    private IPConnection ipcon;
    private BrickMaster  master;
    private BrickDC      motorMaster;
    private BrickServo   servo;

    // Most commonly used bricklets
    private LCD                  lcd;
    private BrickletDualRelay    dualRelay;
    private BrickletIndustrialQuadRelay quadRelay;
    private BrickletDistanceIR   distanceIR;
    private BrickletAmbientLight ambiLight;

    // List of all devices
    private ArrayList<DeviceItem> devices;

    /**
     * Creates a connection to the brick. This may take up to two seconds, the calling
     * thread will be blocked for some time! It is assumed that each stack has at
     * least two devices (master + one other brick or bricklet).
     * 
     * @param host host name or IPv4 address
     * @param port port number (use 4223 as default)
     * @throws IOException if connection fails
     */
    public BrickConnector(String host, int port)
        throws IOException
    {
        devices = new ArrayList<DeviceItem>(10);
        ipcon   = new IPConnection();
        ipcon.setAutoReconnect(false);
        ipcon.addEnumerateListener(this);
        try {
            ipcon.connect(host, port);
            ipcon.enumerate();

            try {
                Thread.sleep(1000);
            }
            catch (InterruptedException i) { /* ignore */ }
            if (devices.size() <= 2) {
                try {
                    Thread.sleep(1000);
                }
                catch (InterruptedException i) { /* ignore */ }
            }
            registerAll();
        }
        catch (Exception e) {
            e.printStackTrace();
            throw new IOException("Connection failed: " + e.getMessage());
        }
    }
    
    /**
     * This method must be called some time after the constructor in order
     * to get feedback from the enumeration which is done in the constructor.
     * <p>
     * Be careful: this methods takes more than a second!
     */
    private void registerAll()
    {
        for (DeviceItem item : devices)
        {
            if (item.getDevice() == null)
                continue;
            
            switch (item.getType()) {
            case MASTER:
                master = (BrickMaster) item.getDevice();
                break;
            case MOTOR:
                motorMaster = (BrickDC) item.getDevice();
                break;
            case SERVO:
                servo = (BrickServo) item.getDevice();
                break;
            case LCD20x4:
                lcd = new LCD((BrickletLCD20x4) item.getDevice());
                break;
            case DUAL_RELAY:
                dualRelay = (BrickletDualRelay) item.getDevice();
                break;
            case QUAD_RELAY:
                quadRelay = (BrickletIndustrialQuadRelay) item.getDevice();
                break;
            case DISTANCE:
                distanceIR = (BrickletDistanceIR) item.getDevice();
                break;
            case LIGHT:
                ambiLight = (BrickletAmbientLight) item.getDevice();
                break;
            default:
                // do nothing, not known
            }
        }
    }

    /* (non-Javadoc)
     * @see com.tinkerforge.IPConnection.EnumerateListener#enumerate(java.lang.String, java.lang.String, short, boolean)
     */
    @Override
    public void enumerate(String uid, String connectedUid, char position, short[] hardwareVersion, short[] firmwareVersion,
                          int deviceIdentifier, short enumerationType)
    {
        if (enumerationType != IPConnection.ENUMERATION_TYPE_AVAILABLE)
            return;

        // Check if UID already exists, should never happen
        if (getDevice(uid) != null)
            System.out.println("Device with UID=" + uid + " is NOT unique! Not added");
        else {
            //System.out.println("UID: " + uid + "  Type-ID: " + deviceIdentifier + "  type: " + enumerationType);
            DeviceItem item = new DeviceItem(ipcon, uid, deviceIdentifier);
            devices.add(item);
        }
    }

    /**
     * Returns one of the master bricks in the stack or null if not found.
     * If more than one master brick is in the stack, the last one found will be returned.
     * @return a device or <code>null</code> if no such device was found
     */
    public BrickMaster getMaster() {
        return master;
    }

    /**
     * Returns one of the DC motor bricks in the stack or null if not found.
     * If more than one dc brick is in the stack, the last one found will be returned.
     * @return a device or <code>null</code> if no such device was found
     */
    public BrickDC getDCMotor() {
        return motorMaster;
    }

    /**
     * Returns one of the LCD 20x4 panels in the stack or null if not found.
     * If more than one device of this type in the stack, the last one found will be returned.
     * @return a device or <code>null</code> if no such device was found
     */
    public LCD getLcd() {
        return lcd;
    }

    /**
     * Returns one of the dual relays in the stack or null if not found.
     * If more than one device of this type in the stack, the last one found will be returned.
     * @return a device or <code>null</code> if no such device was found
     */
    public BrickletDualRelay getDualRelay() {
        return dualRelay;
    }

    /**
     * Returns one of the quad relays in the stack or null if not found.
     * If more than one device of this type in the stack, the last one found will be returned.
     * @return a device or <code>null</code> if no such device was found
     */
    public BrickletIndustrialQuadRelay getQuadRelay() {
        return quadRelay;
    }

    /**
     * Returns one of the ambient light sensors in the stack or null if not found.
     * If more than one device of this type in the stack, the last one found will be returned.
     * @return a device or <code>null</code> if no such device was found
     */
    public BrickletAmbientLight getAmbiLight() {
        return ambiLight;
    }

    /**
     * Returns one of the distance IR sensors in the stack or null if not found.
     * If more than one device of this type in the stack, the last one found will be returned.
     * @return a device or <code>null</code> if no such device was found
     */
    public BrickletDistanceIR getDistanceIR() {
        return distanceIR;
    }

    /**
     * Returns one of the servo bricks in the stack or null if not found.
     * If more than one device of this type in the stack, the last one found will be returned.
     * @return a device or <code>null</code> if no such device was found
     */
    public BrickServo getServo() {
        return servo;
    }

    /**
     * Returns a list of all devices of a given type which are connected.
     * @param type the device type to look for
     * @return a non-null List which maybe empty
     */
    public List<DeviceItem> getDevicesOfType(DeviceType type)
    {
        ArrayList<DeviceItem> result = new ArrayList<DeviceItem>(3);
        for (DeviceItem i : this.devices)
            if (i.getType() == type)
                result.add(i);
        return result;
    }

    /**
     * Returns a device with a given uid.
     * @param uid the device uid to look for
     * @return a device or null if no connected device has such a uid
     */
    public DeviceItem getDevice(String uid)
    {
        for (DeviceItem i : this.devices)
            if (i.hasUid(uid))
                return i;
        return null;
    }
    
    /**
     * Returns the number of devices found.
     * @return an int
     */
    public int getNumDevices() {
        return devices.size();
    }
    
    /**
     * Release all resources, destroy objects.
     * @param openRelais if all relays should be put back into open position
     */
    public void destroy(boolean openRelais)
    {
        if (ipcon == null)
            return;

        try {
            // reset all switches
            if (openRelais)
            {
                for (DeviceItem i : devices)
                {
                    if (i.getType() == DeviceType.QUAD_RELAY)
                        ((BrickletIndustrialQuadRelay) i.getDevice()).setValue(0);
                    if (i.getType() == DeviceType.DUAL_RELAY)
                        ((BrickletDualRelay) i.getDevice()).setState(false, false);
                }
            }
            devices.clear();
            ipcon.disconnect();
        }
        catch (Exception e) {
            e.printStackTrace();
        }
        ipcon = null;
    }
    
    /**
     * TODO
     * 
     * @param sb add the result here
     * @param format a format string which will be used per device in order to add
     *        device specific information to 'sb'
     */
    public void addStatus(StringBuilder sb, String format)
    {
        int len = format.length();
        for (DeviceItem item : this.devices)
        {
            for (int i = 0; i < len; ++i)
            {
                char c = format.charAt(i);
                if (c == '%') {
                    c = (++i < len ? format.charAt(i) :  ' ');
                    switch (c) {
                    case 'N':
                        sb.append(item.getType().toString());
                        break;
                    case 'U':
                        sb.append(item.getUid());
                        break;
                    default:
                        throw new RuntimeException("Invlaid format character '" + c + "'");
                    }
                }
                else
                    sb.append(c);
            }
        }
    }

    @Override
    protected void finalize() {
        destroy(false);
    }
}
