
package de.grosenick.brickapi;

/**
 * Interfaces for complex or simple commands that should be send to the brick
 * over the USB / network. Such commands are typically executed in a separate thread.
 * <P>
 * A command can have multiple status values:<br>
 * <ul>
 *   <li><b>idle</b>: <b>not</b> in the command queue and <b>not</b> executing</li>
 *   <li><b>in queue</b>: in the command queue and <b>not</b> yet executing</li>
 *   <li><b>active</b>: <b>not</b> in the command queue but currently executing</li>
 * </ul>
 * 
 * @author Holger Grosenick
 */
public interface Command extends Cloneable
{
    /**
     * Command status, see {@link Command} description
     */
    public enum Status
    {
        /** <b>not</b> in the command queue and <b>not</b> executing */
        IDLE,
        /** in the command queue and <b>not</b> yet executing */
        IN_QUEUE,
        /** <b>not</b> in the command queue but currently executing */
        ACTIVE
    }

    /**
     * 
     * @param brickConnector the connector to use
     * @throws Exception in case of any error
     */
    void execute(BrickConnector brickConnector)
        throws Exception;

    /**
     * Sets a new status and put this command into the given queue. This is
     * only allowed, if the command is IDLE.
     * @param q queue to put this command to
     * @throws IllegalStateException, if the command is not IDLE.
     */
    void enqueue(java.util.AbstractQueue<Command> q);

    /**
     * Returns a result object (maybe null).
     * @return a result or null
     */
    Object getResult();

    /**
     * Returns true if the command is not in the queue and not active.
     * @return true or false
     */
    boolean isIdle();

    /**
     * Returns true if {@link #enqueue(java.util.AbstractQueue)} was called, but {@link #execute(BrickConnector)}
     * was not yet called: the command is still in the queue.
     * @return true or false
     */
    boolean isInQueue();

    /**
     * Returns true if {@link #execute(BrickConnector)} was started for this object,
     * but has not yet finished.
     * @return true or false
     */
    boolean isActive();
    
    /**
     * Return the current command status
     * @return status object
     * @see #isActive()
     * @see #isIdle()
     * @see #isInQueue()
     */
    Status getStatus();
}
