package de.grosenick.brickapi;

import com.tinkerforge.BrickDC;
import com.tinkerforge.BrickMaster;
import com.tinkerforge.BrickServo;
import com.tinkerforge.BrickletAmbientLight;
import com.tinkerforge.BrickletAnalogIn;
import com.tinkerforge.BrickletDistanceIR;
import com.tinkerforge.BrickletDualRelay;
import com.tinkerforge.BrickletIndustrialQuadRelay;
import com.tinkerforge.BrickletLCD20x4;
import com.tinkerforge.BrickletLCD16x2;
import com.tinkerforge.BrickletLinearPoti;
import com.tinkerforge.BrickletRotaryPoti;
import com.tinkerforge.BrickletTemperature;
import com.tinkerforge.BrickletTemperatureIR;
import com.tinkerforge.BrickletVoltageCurrent;
import com.tinkerforge.Device;
import com.tinkerforge.IPConnection;

/**
 * Generic device that holds a type flag in order to precisely determine the devices.
 * 
 * @author Holger Grosenick
 */
public class DeviceItem
{
    private String uid;
    private Device device;
    private DeviceType type;

    /** enumeration types
    "11", "Brick DC"
    "12", "Brick Debug"
    "13", "Brick Master 2.0"
    "14", "Brick Servo"
    "15", "Brick Stepper"
    "16", "Brick IMU"
    "21", "Bricklet Ambient Light"
    "22", "Bricklet Breakout"
    "23", "Bricklet Current12"
    "24", "Bricklet Current25"
    "25", "Bricklet Distance IR"
    "26", "Bricklet Dual Relay"
    "27", "Bricklet Humidity"
    "28", "Bricklet IO-16"
    "29", "Bricklet IO-4"
    "210", "Bricklet Joystick"
    "211", "Bricklet LCD 16x2"
    "212", "Bricklet LCD 20x4"
    "213", "Bricklet Linear Poti"
    "214", "Bricklet Piezo Buzzer"
    "215", "Bricklet Rotary Poti"
    "216", "Bricklet Temperature"
    "217", "Bricklet Temperature IR"
    "218", "Bricklet Voltage"
    "219", "Bricklet Analog In"
    "220", "Bricklet Analog Out"
    "221", "Bricklet Barometer"
    "222", "Bricklet GPS"
    "223", "Bricklet Industrial Digital In 4"
    "224", "Bricklet Industrial Digital Out 4"
    "225", "Bricklet Industrial Quad Relay"
    "226", "Bricklet PTC"
    "227", "Bricklet Voltage/Current"
    */
    
    /**
     * Create a device item from the values returned by a device enumerator.
     * 
     * @param ipcon IPConnection where to register the device
     * @param uid device id
     * @param deviceType device type
     */
    public DeviceItem(IPConnection ipcon, String uid, int deviceType)
    {
        this.uid    = uid;
        this.device = null;

        switch (deviceType) {
        case BrickDC.DEVICE_IDENTIFIER:
            device = new BrickDC(uid, ipcon);
            type   = DeviceType.MOTOR;
            return;
        case BrickMaster.DEVICE_IDENTIFIER:
            device = new BrickMaster(uid, ipcon);
            type   = DeviceType.MASTER;
            return;
        case BrickServo.DEVICE_IDENTIFIER:
            device = new BrickServo(uid, ipcon);
            type   = DeviceType.SERVO;
            return;

        // Bricklets
        case BrickletAmbientLight.DEVICE_IDENTIFIER:
            device = new BrickletAmbientLight(uid, ipcon);
            type   = DeviceType.LIGHT;
            return;
        case BrickletAnalogIn.DEVICE_IDENTIFIER:
            device = new BrickletAnalogIn(uid, ipcon);
            type   = DeviceType.VOLTAGE;
            return;
        case BrickletDistanceIR.DEVICE_IDENTIFIER:
            device = new BrickletDistanceIR(uid, ipcon);
            type   = DeviceType.DISTANCE;
            return;
        case BrickletDualRelay.DEVICE_IDENTIFIER:
            device = new BrickletDualRelay(uid, ipcon);
            type   = DeviceType.DUAL_RELAY;
            return;
        case BrickletLCD16x2.DEVICE_IDENTIFIER:
            device = new BrickletLCD16x2(uid, ipcon);
            type   = DeviceType.LCD20x4;
            return;
        case BrickletLCD20x4.DEVICE_IDENTIFIER:
            device = new BrickletLCD20x4(uid, ipcon);
            type   = DeviceType.LCD20x4;
            return;
        case BrickletLinearPoti.DEVICE_IDENTIFIER:
            device = new BrickletLinearPoti(uid, ipcon);
            type   = DeviceType.LINEAR_POTI;
            return;
        case BrickletIndustrialQuadRelay.DEVICE_IDENTIFIER:
            device = new BrickletIndustrialQuadRelay(uid, ipcon);
            type   = DeviceType.QUAD_RELAY;
            return;
        case BrickletRotaryPoti.DEVICE_IDENTIFIER:
            device = new BrickletRotaryPoti(uid, ipcon);
            type   = DeviceType.ROTARY_POTI;
            return;
        case BrickletTemperature.DEVICE_IDENTIFIER:
            device = new BrickletVoltageCurrent(uid, ipcon);
            type   = DeviceType.TEMPERATURE;
            return;
        case BrickletTemperatureIR.DEVICE_IDENTIFIER:
            device = new BrickletTemperatureIR(uid, ipcon);
            type   = DeviceType.TEMPERATURE_IR;
            return;
        case BrickletVoltageCurrent.DEVICE_IDENTIFIER:
            device = new BrickletVoltageCurrent(uid, ipcon);
            type   = DeviceType.VOLTAGE_CURRENT;
            return;
        }

        System.err.println("Unknow device: type=" + deviceType + " UID=" + uid);
        type = DeviceType.UNKNOWN;
    }

    /**
     * Returns the device id.
     * @return non-null string
     */
    public String getUid() {
        return uid;
    }

    /**
     * Checks if the devices has the given uid.
     * @param id the uid to check
     * @return true if equal, false if not.
     */
    public boolean hasUid(String id) {
        return uid.equals(id);
    }

    /**
     * Returns the device object-
     * @return non-null object pointer
     */
    public Device getDevice() {
        return device;
    }

    /**
     * Returns the device type.
     * @return type flag
     */
    public DeviceType getType() {
        return type;
    }
}