
package de.grosenick.brickapi;

import com.tinkerforge.BrickletLCD20x4;

/**
 * LCD class that has some more text features for the LCD screen and a functionality
 * that button 1 switches the backlight on / off.
 * 
 * @author Holger Grosenick
 */
public class LCD
{
    private final short maxCols;
    private final short maxLines;

    private BrickletLCD20x4 lcd;
    private short    line;
    private String[] text;
    private boolean  isOn = false;
    private boolean  button1Light;

    /**
     * Helper class that implements the bricklet callback for the LCD buttons.
     * By default it turns backlight on/off by pressing the first button.
     */
    private class ButtonListener implements BrickletLCD20x4.ButtonPressedListener
    {
        @Override
        public void buttonPressed(short button)
        {
            if (button1Light && button == 0) {
                toggleBacklight();
            }
        }
    }

    /**
     * Init the LCD with a 20x4 LCD display.
     * @param lcd the display to use
     */
    public LCD(BrickletLCD20x4 lcd)
    {
        this.maxLines = 4;
        this.maxCols  = 20;
        this.lcd      = lcd;
        this.text     = new String[maxLines];
        this.line     = 0;
    }
    
    /**
     * Enables the function that button1 (1 == first of 4) switches the light on or off.
     * This feature cannot be disabled afterwards. It should not be used together with
     * {@link #addButtonPressListener(com.tinkerforge.BrickletLCD20x4.ButtonPressedListener)},
     * these features might overlap.
     */
    public void enableLightOnButton1() {
        addButtonPressListener(new ButtonListener());
    }

    /**
     * Adds a button press listener which is outside this class.
     * @param l the listener to use
     * @see #enableLightOnButton1()
     */
    public void addButtonPressListener(BrickletLCD20x4.ButtonPressedListener l)
    {
        if (!button1Light) {
            button1Light = true;
            lcd.addButtonPressedListener(l);
        }
    }

    /**
     * Clears the display content.
     */
    public void clearDisplay() {
        try {
            lcd.clearDisplay();
        }
        catch (Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * Switches the backlight on.
     */
    public void backlightOn() {
        try {
            lcd.backlightOn();
        }
        catch (Exception e) {
            e.printStackTrace();
            return;
        }
        isOn = true;
    }

    /**
     * Switches the backlight off.
     */
    public void backlightOff() {
        try {
        lcd.backlightOff();
        }
        catch (Exception e) {
            e.printStackTrace();
            return;
        }
        isOn = false;
    }
    
    /**
     * Toggles the backlight to on / off depending on the current state.
     * @return the new backlight status: true (on) or false (off)
     */
    public boolean toggleBacklight() {
        setBacklight(!isOn);
        return isOn;
    }
    
    /**
     * Switches the backlight on or off depending on the argument.
     * @param on true (ON) or false (OFF)
     */
    public void setBacklight(boolean on) {
        if (on) {
            backlightOn();
        }
        else {
            backlightOff();
        }
    }
    
    /**
     * Returns true if the LCD backlight is on, this method re-reads
     * the effective status from the device, for those cases where the
     * initial state is unknown.
     * 
     * @return true if the backlight is on
     */
    public boolean isBacklightOn() {
        try {
            isOn = lcd.isBacklightOn();
            return isOn;
        }
        catch (Exception e) {
            e.printStackTrace();
            return false;
        }
    }

    /**
     * Clears the whole display and fills it with the given text starting at position (0,0).
     * If the text has a newline, a new line is started.
     * 
     * @param msg the text to display on the screen.
     */
    public void fillWith(String msg)
    {
        String[] lines = msg.split("\n");
        clearDisplay();
        
        line = 0;
        for (short i = 0; i < lines.length; ++i)
        {
            String l = lines[i];
            addLine(l);
        }
    }

    /**
     * Adds a line to the current position and increments the line counter.
     * If the line counter has reached the max number of lines, the whole
     * contents if scrolled up.
     * @param t String to add to the next line.
     */
    public void addLine(String t)
    {
        try {
            if (line+1 >= maxLines) {
                // scroll
                for (short i = 0; i+1 < maxLines; ++i) {
                    text[i] = text[i + 1];
                    lcd.writeLine(i, (short) 0, text[i]);
                }
            }
        
            if (t.length() >= maxCols) {
                text[line] = t.substring(0, maxCols);
            }
            else {
                StringBuilder s = new StringBuilder(20);
                s.append(t);
                while (s.length() < 20)
                    s.append(' ');
                text[line] = s.toString();
            }
            lcd.writeLine(line, (short) 0, text[line]);
        }
        catch (Exception e) {
            e.printStackTrace();
        }
        
        if (line+1 < maxLines)
            ++line;
    }
}
