
package de.grosenick.brickapi.commands;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;

import com.tinkerforge.BrickServo;
import com.tinkerforge.BrickletDistanceIR;

import de.grosenick.brickapi.AbstractCommand;
import de.grosenick.brickapi.BrickConnector;

/**
 * Command that changes the state of a servo. The command also has the function
 * to be controlled by a distance sensor: if the distance gets below a certain
 * value, the servo movement should stop and allow only values into the other
 * direction.
 * 
 * @author Holger Grosenick
 */
public class MoveServo extends AbstractCommand implements BrickletDistanceIR.DistanceReachedListener
{
    /** 
     * The number of values which are expected in a value array which can be set using
     * {@link #setDegreeValues(int[])}, the array has one value for the percent range
     * <code>-100 .. 0 .. 100</code>.
     */
    public static final int NUM_SERVO_POINTS = 201;
    
    /**
     * Created if the distanceReachedCallback is created the first time.
     */
    private class AutoStop
    {
        AutoStop()
        {
            created = System.currentTimeMillis();
            lastHappend = created;
            minDistancePercent = percentValue;
        }

        long  created;
        long  lastHappend;
        int   minDistancePercent;
    }

    private short   servoNum;
    private short   servoNum2 = -1;
    private short   position;
    private short   minDistance;         // set if combined with distance IR
    private int     percentValue;        // current percentage of the max degree value (-100 .. +100)
    private int     maxPercent = 100;    // max percentage of the max degree value (20 .. 100)
    private int     degrees;             // max degree value
    private float   degreesPerPercent;
    private boolean init;
    private boolean disabled;
    private boolean negate           = false;
    private boolean negate2          = false;
    private boolean shouldSnap       = true;
    private boolean shouldSnapCenter = true;
    private boolean changed          = true;
    private String  label            = "no name";
    private int[]   degreeValues     = null;
    private BrickServo myServo;
    private AutoStop   autoStop;
    
    /**
     * The result for {@link MoveServo#loadFile(String, String)}.
     */
    public static class LoadFileResult
    {
        /** Return status: OK or any kind of error? */
        public enum Result {
            /** OK, data loaded 'data' is set */
            OK,
            /** File has not enough items */
            INSUFFICIENT_ITEMS,
            /** one line is invalid, the contents is not a number */
            NUMBER_ERROR,
            /** any other error */
            ERROR
        };
        /** main status code */
        public Result result;
        /** optional parameter in case of error */
        public String errorParameter;
        /** data in case of result == OK */
        public int[] data;
    }

    /**
     * Init the servo with initial commands and set the max range for the values.
     * This object should be reused to control the servo as long as possible.
     * <P>
     * This constructor will use the servo in linear mode and normal direction
     * (not inverted).
     * 
     * @param servoNum value 1..7 - other values disable the servo!
     * @param degrees rotation range in degrees (10 .. 180)
     * @param init if set to true, {@link #doExecute(BrickConnector)} will just
     *        init the servo and then set the flag to false.
     */
    public MoveServo(int servoNum, int degrees, boolean init)
    {
        if (degrees < 10 || degrees > 180)
            throw new IllegalArgumentException("'degrees' value must be between 10..180");
        if (servoNum < 0 || servoNum > 7)
            throw new IllegalArgumentException("'servoNum' value must be between 0..7");
        this.disabled = servoNum <= 0 || servoNum > 7;
        this.servoNum = (short) (servoNum-1);
        this.degrees  = degrees;
        this.init     = init;
        degreesPerPercent = degrees * maxPercent / 100;
    }

    /**
     * Loads the given file as file with numbers, each line has one number which is a point.
     * 
     * @param dir directory where to find the file
     * @param filename file to load
     * @return a array of int or null in case of error
     */
    public static LoadFileResult loadFile(String dir, String filename)
    {
        File f = new File(dir, filename);
        FileReader is = null;
        int itemNo = 0;
        String line = null;
        
        LoadFileResult res = new LoadFileResult();
    
        try {
            // open file
            is = new FileReader(f);
            BufferedReader rd = new BufferedReader(is);
    
            // read all lines
            int[] newValues = new int[NUM_SERVO_POINTS];
    
            // read lines: one number per line
            while ((line = rd.readLine()) != null)
            {
                line = line.trim();
                if (line.length() == 0 || line.charAt(0) == '#')
                    continue;
                if (itemNo == newValues.length) {
                    // skip the rest of the file
                    break;
                }
                int x = Integer.parseInt(line);
                if (x < -100 || x > 100)
                    throw new NumberFormatException("out of range");
                newValues[itemNo] = x;
                ++itemNo;
            }
            if (itemNo != newValues.length) {
                res.result = LoadFileResult.Result.INSUFFICIENT_ITEMS;
            }
            else {
                res.result = LoadFileResult.Result.OK;
                res.data = newValues;
            }
        }
        catch (NumberFormatException io) {
            res.result = LoadFileResult.Result.NUMBER_ERROR;
            res.errorParameter = line;
        }
        catch (Exception io) {
            res.result = LoadFileResult.Result.ERROR;
            res.errorParameter = io.getMessage();
        }
        finally {
            if (is != null) {
                try {
                    is.close();
                }
                catch (IOException e) { /* ignore */ }
            }
        }
        return res;
    }

    /**
     * Sets or unsets the negation of the servo direction.
     * @param negate negate servo direction?
     */
    public void setNetgate(boolean negate) {
        this.negate = negate;
    }
    
    /**
     * 
     * @param channel
     * @param negate negate servo direction?
     */
    public void setServo2(int channel, boolean negate)
    {
        this.servoNum2 = (short)(channel-1);
        this.negate2   = negate;
    }
    
    /**
     * Sets multiple servo options at once.
     * 
     * @param negate negate servo direction?
     * @param snap set servo back to 0 on release?
     * @param snapCenter set servo back to 0 on release near to the center?
     * @param maxPercent 100% of degree value equals this value of max servo range
     * @see #shouldSnap()
     * @see #setNetgate(boolean)
     */
    public void setOptions(boolean negate, boolean snap, boolean snapCenter, int maxPercent)
    {
        this.negate     = negate;
        this.shouldSnap = snap;
        this.shouldSnapCenter = snapCenter;
        setMaxPercent(maxPercent);
    }
    
    /**
     * Sets a stop distance in cm: if this distance is reached, the servo is stopped
     * and only movements to the other direction are possible.
     * @param cm distance in cm
     */
    public void stopOnDistance(int cm)
    {
        // value is in millimeters
        minDistance = (short) (cm * 10);
    }
    
    /**
     * Returns true if the channel was set to 0 which would disable this control.
     * @return true or false
     */
    public boolean isDisabled() {
        return disabled;
    }
    
    /**
     * Returns the max range (in %) of the servo range/movement. A value of 50 means
     * that <code>rotateTo(100)</code> will rotate to 100% of the maximum which is
     * set to 50. So the servo will move to 50% of its max range.
     * @return an int
     */
    public int getMaxPercent() {
        return maxPercent;
    }
    
    /**
     * Limits the servo movement to X% of its max range.
     * @param maxPercent 100% of degree value equals this value of max servo range
     * @see #getMaxPercent()
     */
    public void setMaxPercent(int maxPercent) {
        if (maxPercent >= 20 && maxPercent <= 100) {
            this.maxPercent = maxPercent;
            degreesPerPercent = degrees * maxPercent / 100;
        }
    }
    
    /**
     * Sets a new array with percent value mappings: each entry of the array maps
     * the percent value P into <code>newValues[P + 100]</code> where <code>P = -100 .. 100</code>.
     * @param newValues
     */
    public void setDegreeValues(int[] newValues) {
        this.degreeValues = newValues;
    }
    
    /**
     * Calculates a new servo position, the result is specific for that servo
     * and defines a controller value.
     * 
     * @param percent the percentage of the max degree value to move to
     * @return servo movement value
     */
    public short calculatePosition(int percent)
    {
        short result;

        if (degreeValues != null) {
            if (percent >= -100 && percent <= 100)
                percent = degreeValues[percent+100];
            else
                throw new RuntimeException("Invalid percent value " + percent);
        }

        // linear control if degreeValues == 0
        if (negate)
            result = (short) (degreesPerPercent * -percent);
        else
            result = (short) (degreesPerPercent * percent);
        return result;
    }

    /* (non-Javadoc)
     * @see de.grosenick.brickapi.AbstractCommand#doExecute(de.grosenick.brickapi.BrickConnector)
     */
    @Override
    protected void doExecute(BrickConnector brickConnector) throws Exception
    {
        if (changed == false)
            return;

        BrickServo servo = brickConnector.getServo();
        if (disabled || servo == null)
            return;
        
        if (init)
        {
            servo.enable(servoNum);
            servo.setOutputVoltage(5900);
            servo.setVelocity(servoNum, 65535);
            servo.setAcceleration(servoNum, 65535);
            //servo.setPeriod(servoNum, 19600);
            servo.setDegree(servoNum, (short) (degrees * -100), (short) (degrees * 100));

            if (servoNum2 >= 0)
            {
                servo.enable(servoNum2);
                servo.setOutputVoltage(5900);
                servo.setVelocity(servoNum2, 65535);
                servo.setAcceleration(servoNum2, 65535);
                servo.setDegree(servoNum2, (short) (degrees * -100), (short) (degrees * 100));
            }

            percentValue = 0;
            init = false;
            position = calculatePosition(0);
            myServo  = servo;
            
            BrickletDistanceIR ir;
            if (minDistance > 0 && (ir = brickConnector.getDistanceIR()) != null)
            {
                ir.setDebouncePeriod(500);
                ir.setDistanceCallbackThreshold('<', minDistance, (short)0);
                ir.addDistanceReachedListener(this);
                //System.err.println("Set callback for distance " + minDistance);
            }
        }
        //System.out.println("Servo " + servoNum + ' ' + position);
        servo.setPosition(servoNum, position);
        if (servoNum2 >= 0) {
            //System.out.println("Servo " + servoNum2 + ' ' + position);
            servo.setPosition(servoNum2, negate != negate2 ? (short)-position : position);
        }
        changed = false;
    }
    
    /**
     * Returns a servo label (non-null).
     * @return a string
     */
    public String getLabel() {
        return label;
    }

    /**
     * 
     * @return the actual position in percent of the max range (-100 .. 100)
     */
    public int getPercentValue() {
        return percentValue;
    }
    
    /**
     * Returns true if the last call {@link #rotateTo(int)} has changed the position
     * of the servo and a new command must be transfered over network.
     * @return true or false
     * @see #forceUpdate()
     */
    public boolean isChanged() {
        return changed;
    }
    
    /**
     * Sets the 'changed' flag which indicates that the current position must be
     * transfered over network again.
     */
    public void forceUpdate() {
        changed = true;
    }

    /**
     * Sets the percentage (-100 - +100) of the max angle to rotate the servo to.
     * If the autoStop function is active, specific movements might be disabled
     * in that time.
     * 
     * @param percentage percentage of the max servo angle to move to
     * @see #distanceReached(int)
     */
    public void rotateTo(int percentage)
    {
        // System.out.println("Rotate to old=" + percentValue + " new=" + percentage);
        if (autoStop != null)
        {
            long current = System.currentTimeMillis();
            if ((current - autoStop.lastHappend) >= 1500)
                // if elapsed time since last rebounce is more than a second: clear
                autoStop = null;
            else {
                // if the movement is still in the same direction: do not drive
                if ((percentage < 0 && autoStop.minDistancePercent < 0) ||
                    (percentage > 0 && autoStop.minDistancePercent > 0))
                    return;
            }
        }

        percentValue = percentage;
        short p = calculatePosition(percentage);
        if (p == position)
            return;

        // really update
        position = p;
        changed = true;
    }
    
    /**
     * Changes the servo label.
     * @param lbl a non-null string
     */
    public void setLabel(String lbl) {
        if (lbl != null)
            label = lbl;
    }
    
    /**
     * Returns true if this servo should snap back to the center if the
     * control is released by the user (default is 'true').
     * @return true or false
     */
    public boolean shouldSnap() {
        return shouldSnap;
    }
    
    /**
     * Returns true if this servo should snap back to the center if the
     * control is released <b>nearby the center</b> (default is 'true').
     * @return true or false
     */
    public boolean shouldSnapCenter() {
        return shouldSnapCenter;
    }
    
    /**
     * Callback method if the distance gets below a certain point. In this case,
     * the servo should stop! {@link #rotateTo(int)} will allow only movements in
     * the other direction then. If the distance callback is not called for more
     * than a second, the stop-function gets cleared.
     */
    @Override
    public void distanceReached(int distance)
    {
        // System.err.println("Distance-callback STOP at call #" + callCount);
        if (autoStop != null)
        {
            autoStop.lastHappend = System.currentTimeMillis();
            return;
        }

        // something else crossed the sensor
        if (position == 0)
            return;

        // New event: move slowly in the other direction.
        // Some DC engine controllers make a full stop if the
        // direction is changed abruptly
        autoStop = new AutoStop();
        if (autoStop.minDistancePercent < 0)
            rotateTo(15);
        else
            rotateTo(-15);

        try {
            if (myServo != null) {
                myServo.setPosition(servoNum, position);
            }
        }
        catch (Exception e) {
            myServo = null;
        }
    }
}
