
package de.grosenick.brickapi.commands;

import com.tinkerforge.BrickMaster;
import com.tinkerforge.BrickMaster.WifiBufferInfo;
import com.tinkerforge.BrickMaster.WifiStatus;
import com.tinkerforge.BrickletAmbientLight;

import de.grosenick.brickapi.AbstractCommand;
import de.grosenick.brickapi.BrickConnector;

/**
 * Command that reads the current voltage and light sensor out of a stack
 * and creates a status string out of that. This commands also calculates
 * a history of illuminance values which can be used to automatically
 * trigger a light function.
 * 
 * @author Holger Grosenick
 */
public class ReadStatus extends AbstractCommand
{
    /** Max number of sensor values put into an internal list, used for {@link #getIlluminanceAverage(int)} */
    public static int MAX_STATUS_VALUES = 15;

    private int stackMilliVolt = -1;
    private int stackMilliAmp  = -1;
    private int illuminance    = -1;
    private int[] illuminances;
    private int[] voltages;

    private boolean        current;
    private boolean        updateWifiStatus;
    private WifiStatus     wifiStatus;
    private WifiBufferInfo wifiBufferInfo;
    private StringBuilder  myStringBuilder = new StringBuilder(50);
    private String         myStatus;
    
    /** for automatic light control and power control */
    private int minLuxOn;
    private int minVoltage;

    /**
     * Constructor that will query later on all attached devices.
     */
    public ReadStatus() {
        current = true;
        illuminances = makeArray();
        voltages     = makeArray();
    }

    /**
     * Constructor that may suppress reading some sensor values.
     * 
     * @param voltage read voltage of stack?
     * @param current read current of stack?
     * @param light read light sensor too?
     */
    public ReadStatus(boolean voltage, boolean current, boolean light)
    {
        this.current = current;
        if (light)
            illuminances = makeArray();
        if (voltage)
            voltages = makeArray();
    }

    /**
     * Create an array of int of size MAX_STATUS_VALUES.
     */
    private int[] makeArray()
    {
        int[] res = new int[MAX_STATUS_VALUES];
        for (int i = 0; i < MAX_STATUS_VALUES; ++i)
            res[i] = -1;
        return res;
    }

    /* (non-Javadoc)
     * @see de.grosenick.brickapi.AbstractCommand#doExecute(de.grosenick.brickapi.BrickConnector)
     */
    @Override
    protected void doExecute(BrickConnector brickConnector) throws Exception
    {
        BrickMaster master = brickConnector.getMaster();
        if (master != null) {
            if (voltages != null) {
                System.arraycopy(voltages, 0, voltages, 1, MAX_STATUS_VALUES-1);
                stackMilliVolt = master.getStackVoltage();
                voltages[0] = stackMilliVolt;
            }
            if (current)
                stackMilliAmp  = master.getStackCurrent();
            if (updateWifiStatus) {
                master.refreshWifiStatus();
                wifiStatus = master.getWifiStatus();
                wifiBufferInfo = master.getWifiBufferInfo();
            }
        }

        // build new status string
        StringBuilder sb = myStringBuilder;
        if (getStackMilliVolt() >= 0)
            sb.append(getStackMilliVolt()).append(" mV");
        if (getStackMilliAmp() >= 0) {
            if (sb.length() > 0)
                sb.append(" / ");
            sb.append(getStackMilliAmp()).append(" mA");
        }

        // read the light sensor if request + available
        if (illuminances != null)
        {
            BrickletAmbientLight lightSens = brickConnector.getAmbiLight();
            if (lightSens != null) {
                System.arraycopy(illuminances, 0, illuminances, 1, MAX_STATUS_VALUES-1);
                illuminance = lightSens.getIlluminance();
                illuminances[0] = illuminance;

                // Display the average illuminance of the last seconds
                int avg = (int) getIlluminanceAverage(3);
                if (avg >= 0) {
                    if (sb.length() > 0)
                        sb.append(" / ");
                    sb.append(avg).append(" Lux");
                }
            }
            else
                illuminance = -1;

        }

        myStatus = sb.toString();
        sb.delete(0, sb.length());
    }

    /**
     * Returns the current stack voltage in mV.
     * @return an int or -1 if no master was found.
     */
    public int getStackMilliVolt() {
        return stackMilliVolt;
    }

    /**
     * Returns the current stack current usage in mA.
     * @return an int or -1 if no master was found.
     */
    public int getStackMilliAmp() {
        return stackMilliAmp;
    }

    /**
     * Returns the current value of the light sensor.
     * @return an int or -1 if no light sensor was found.
     */
    public int getIlluminance() {
        return illuminance;
    }

    /**
     * Builds the average value out of the latest <i>count</i> values of the light sensor.
     * This makes only sense if this object is pushed onto the command stack more than once.
     * 
     * @param count number of value to use in order to calculate the average
     * @return the average value
     */
    public float getIlluminanceAverage(int count)
    {
        // the value is Lux * 10, cut off decimals
        return getAverage(illuminances, count) / 10;
    }
    
    /**
     * Sets a flag so that the next status update will refresh the Wifi status too, as
     * long as the status update wasn't executed, {@link #doUpdateWifiStatus()} will
     * return null.
     * @see #getLatestWifiStatus()
     */
    public void doUpdateWifiStatus() {
        this.updateWifiStatus = true;
        this.wifiStatus = null;
        this.wifiBufferInfo = null;
    }
    
    /**
     * Returns the latest WifiStatus if available, the status is updated <b>once</b>
     * after {@link #doUpdateWifiStatus()} is called.
     * @return an object or null.
     */
    public WifiStatus getLatestWifiStatus() {
        return wifiStatus;
    }
    
    /**
     * Returns the latest WifiBufferInfo if available, the status is updated <b>once</b>
     * after {@link #doUpdateWifiStatus()} is called.
     * @return an object or null.
     */
    public WifiBufferInfo getLatestWifiBufferInfo() {
        return wifiBufferInfo;
    }

    /**
     * Determines the minimum illuminance value out of the latest <i>count</i> values
     * of the light sensor.
     * 
     * @param count number of value to use in order to find the minimum
     * @return the average value in Lux*10
     */
    public int getIlluminanceMin(int count)
    {
        if (illuminances == null)
            return -1;
        int min = illuminances[0];
        int last = count > illuminances.length ? illuminances.length : count;
        if (last > callCount)
            last = callCount;

        for (int i = 1; i < last; ++i) {
            if (illuminances[i] < min);
                min = illuminances[i];
        }
        return min;
    }

    /**
     * Determines the maximum illuminance value out of the latest <i>count</i> values
     * of the light sensor.
     * 
     * @param count number of value to use in order to find the maximum
     * @return the average value in Lux*10
     */
    public int getIlluminanceMax(int count)
    {
        if (illuminances == null)
            return -1;
        int max = illuminances[0];
        int last = count > illuminances.length ? illuminances.length : count;
        if (last > callCount)
            last = callCount;

        for (int i = 1; i < last; ++i) {
            if (illuminances[i] > max);
                max = illuminances[i];
        }
        return max;
    }

    /**
     * Returns a string with the latest status values.
     * @return a string
     */
    public String getStatusString() {
        return myStatus;
    }

    /**
     * Builds the average value out of the latest <i>count</i> values of voltage measurement.
     * This makes only sense if this object is pushed onto the command stack more than once.
     * 
     * @param count number of value to use in order to calculate the average
     * @return the average value
     */
    public float getVoltageAverage(int count)
    {
        return getAverage(voltages, count);
    }
    
    /**
     * Builds an average of 'count' values out of the given value list.
     * @return the average
     */
    private float getAverage(int[] values, int count)
    {
        if (values == null)
            return -1;
        int last = count > values.length ? values.length : count;
        if (last > callCount)
            last = callCount;

        int avg = 0;
        int i = 0;

        //StringBuilder sb = (new StringBuilder(200)).append("Call #").append(callCount).append(" - add");
        while (i < last) {
            //sb.append(' ').append(values[i]);
            avg += values[i];
            ++i;
        }
        float result = i == 0 ? 0 : (float) avg / (float) i;
        //System.out.println(sb.append(" = ").append(result).toString());
        return result;
    }

    /**
     * Returns true if the last light value if below the value of 'minLuxOn',
     * returns false if not light sensor is present or minLuxOn is 0.
     * 
     * @return true or false
     * @see #setMinLuxOn(int)
     * @see #isEnoughLight()
     */
    public boolean isTooDark()
    {
        if (illuminances == null || minLuxOn <= 0)
            // 'isEnoughLight' always returns true, but this always false => consistent
            return false;

        // only the last value
        int maxLux = getIlluminanceMax(1);
        
        // currently dark or enough light (less than 15 Lux)
        return maxLux >= 0 && maxLux < minLuxOn;
    }
    
    /**
     * Returns true if the average light value of the <b>last 3</b> measurements
     * is <b>above</b> minLuxOn or no light sensor if available. Returns false if
     * the light value is larger than 0 but below or equal to min LuxOn.
     * 
     * @return true or false
     * @see #setMinLuxOn(int)
     * @see #isTooDark()
     */
    public boolean isEnoughLight()
    {
        if (illuminances == null || minLuxOn <= 0)
            // 'isTooDark' always returns false, but this always true => consistent
            return true;

        // enough light in the last 3 periods: switch light off
        return getIlluminanceAverage(3) >= (minLuxOn + 10);
    }
    
    /**
     * Returns true if the average of the last 4 voltage measurements was is below the
     * minimum value defined by {@link #setMinVoltage(int)}. If voltage is
     * disabled, the result is always false.
     * 
     * @return true or false.
     */
    public boolean isLowVoltage() {
        return voltages != null && minVoltage > 0 && this.getVoltageAverage(5) < minVoltage;
    }

    /**
     * Registers a min light value in Lux when the light should go on.
     * An external user must query 
     * @param minLuxOn min value in Lux, a value of 0 disables the functionality
     */
    public void setMinLuxOn(int minLuxOn) {
        this.minLuxOn = minLuxOn * 10;
    }
    
    /**
     * Sets a minimum voltage to check in {@link #isLowVoltage()}.
     * 
     * @param mV value in millivolts
     * @see #isLowVoltage()
     */
    public void setMinVoltage(int mV) {
        this.minVoltage = mV;
    }
}
