
package de.grosenick.brickapi.commands;

import com.tinkerforge.BrickletIndustrialQuadRelay;

import de.grosenick.brickapi.AbstractCommand;
import de.grosenick.brickapi.BrickConnector;
import de.grosenick.brickapi.DeviceItem;
import de.grosenick.brickapi.DeviceType;

/**
 * Command that changes the state of a quad relay. It offers monoflop
 * and normal switch operations.
 * 
 * @author Holger Grosenick
 */
public class SetRelay extends AbstractCommand
{
    private final int relayNum;
    private final int myBit;

    private int     duration;
    private boolean on;
    private String  uid;

    /**
     * Init relay command just with the switch number.
     * 
     * @param uid optional relay-uid, if set, this relay is searched
     * @param relayNum value 0..3
     */
    public SetRelay(String uid, int relayNum)
    {
        if (relayNum < 0 || relayNum > 3)
            throw new IllegalArgumentException("'relayNum' value must be between 0..3");
        this.duration = -1;
        this.myBit    = 1 << relayNum;
        this.relayNum = relayNum;
        this.debugOn  = false;
        if (uid != null && uid.length() >= 3)
            this.uid = uid;
    }

    /* (non-Javadoc)
     * @see de.grosenick.brickapi.AbstractCommand#doExecute(de.grosenick.brickapi.BrickConnector)
     */
    @Override
    protected void doExecute(BrickConnector brickConnector) throws Exception
    {
        BrickletIndustrialQuadRelay relay = null;
        if (uid == null)
            relay = brickConnector.getQuadRelay();
        else {
            DeviceItem item = brickConnector.getDevice(uid);
            if (item != null && item.getType() == DeviceType.QUAD_RELAY) {
                relay = (BrickletIndustrialQuadRelay) item.getDevice();
            }
        }
        if (relay == null) {
            if (debugOn)
                System.err.println("NO QUAD RELAY WITH uid=" + uid);
            return;
        }

        if (duration > 0 && on) {
            if (debugOn)
                System.err.println("QUAD MONOFLOP " + myBit + " for " + duration + "ms");
            relay.setMonoflop(myBit, myBit, duration);
            return;
        }

        relay.setSelectedValues(myBit, on ? myBit : 0);
        if (debugOn)
            System.err.println("UPD QUAD RELAY -> " + (on ? "on" : "off"));
    }

    /**
     * Switches the relay on (connected), the monoflop durations stays unchanged,
     * so if the duration is set, the next switch action will be a monoflop.
     */
    public void switchOn() {
        on = true;
    }
    
    /**
     * Sets a duration in milliseconds for which the switch will be on, it
     * goes out automatically after that period.
     * @param milliseconds time value in milliseconds
     */
    public void setDuration(int milliseconds) {
        this.duration = milliseconds;
    }

    /**
     * Switches the relay on for a specific amount of time, the goes
     * is turned off automatically if the time elapses.
     * 
     * @param durationInMillis monoflop duration in milliseconds
     */
    public void switchOn(int durationInMillis) {
        on = true;
        duration = durationInMillis;
    }
    
    /**
     * Toggle the switch to the specified value.
     * 
     * @param b set switch on or off (false)
     */
    public void switchTo(boolean b) {
        this.on = b;
    }

    /**
     * Switches the relay off (disconnected).
     */
    public void switchOff() {
        on = false;
    }

    /**
     * Returns the duration in ms for which will be switched on (Mono-flop mechanism),
     * a value less than zero shows, that the switch is normally switched on/off and
     * does not turn off automatically.
     * @return and int
     */
    public int getDuration() {
        return duration;
    }

    /**
     * Returns the switch number of the relay.
     * @return number of the relay switch to control
     */
    public int getRelayNum() {
        return relayNum;
    }
    
    /**
     * Returns the UID that the relay should be connected to.
     * The value can be null
     * @return a string or null
     */
    public String getUid() {
        return uid;
    }
    
    /**
     * Updates the UID of the relay to connect to.
     * @param uid new uid or null
     */
    public void setUid(String uid) {
        this.uid = uid;
    }
}
